#include <stdio.h>
#include "lcd_lib.h"
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include "gpscoords.h"
#include <inttypes.h>
#include "uart_test.h"
#include <math.h>
#include "uart.h"


//timeout counter for lcd, led       
volatile uint16_t count_lcd;   
volatile uint16_t count_led;
volatile uint16_t count_adc;
volatile uint16_t count_but;
//time for task
#define time_lcd 500
#define time_led 500
#define time_adc 100 
#define time_but 50 

//State machine state names
#define NoPush 1 
#define MaybePush 2
#define Pushed 3
#define MaybeNoPush 4

//Menu state names
#define MenuCoord 1
#define MenuDiff 2
#define MenuNavig 3
#define MenuTransmit 4
#define MenuMode 5

//Mode state names
#define ModeExplorer 1
#define ModeNavigator 2

unsigned char PushFlag = 0;			//message indicating a button push 
unsigned char PushState = NoPush;	//state machine
unsigned char MenuState = MenuMode;  
unsigned char mode = ModeNavigator;


int current_coord_index = 0;
int explorer_coord_index = 0;
#define NUM_EXPLORER_WAYPTS 100
float lat_exp[NUM_EXPLORER_WAYPTS];
float lon_exp[NUM_EXPLORER_WAYPTS];

void receive_gps_uart(char);

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);



signed int ADC_num;
char ADC_str[] = "    "; // size 4

// heading variables
#define MIN_DIST_WAY 0.00015 //in degrees
#define MIN_PROG 0.000025
#define MIN_HEAD_DIFF 15
#define PI 3.14159265
float prevLat = 0;
float prevLon = 0;
float curHeading = 39;
float forget = 0; // newHeading = forget*oldHeading + (1-forget)*newHeadingReading
char nearWaypoint=0;

// ledblink variables
#define CHIRP_RPT 3000
unsigned char num_syll= 3;
unsigned int syll_length= 100;
unsigned int chirp_rpt_count = 0;
unsigned char num_syll_count = 0;
unsigned int syll_length_count = 0;
char directionWest = 0; //blink left if 1

char gpschar[12];
float debug_count = 0.777;
char debug_str[5];


char gps_count = 0;
char switch_bit = 0x08;

float distLat = 24;
float distLon = 35;
float newHeading = 123;
float diffHead = 111;
float curLatg = 12;
float curLong = 34;
float tarHeading = 256;

// the usual
#define begin {
#define end }

void initialize();
void sendToComputer(void);
float convertGPS(float);
char closeCoords(float, float, float, float, float);
void ledblink(void);
void updateHeading(float, float);


//**********************************************************
//millisecond counter (timer 0) overflow ISR
ISR (TIMER0_COMPA_vect) // every 1ms
begin   

  //Decrement the time if not already zero
  if (count_lcd>0)	--count_lcd;
  if (count_led>0)	--count_led;
  if (count_adc>0)  --count_adc;
  if (count_but>0)  --count_but;
  if (nearWaypoint) ledblink();
end  
 
char rx_c[1024];
int rx_size = 0;
char sentence[1024];
int sentence_len = 0;

  // -------  LCD Printout  ----------------------------------------
void LCD_print() 
  begin

	char line2[10];

	float curLat_t;
	float curLon_t;
	char type[6];
	int quality = 0;
	char valid; // A=active, V=void
	float time;


	
	sscanf(sentence, "%5c,%f,%c,%f,%*c,%f,%*c,%d", type, &time, &valid,&curLat_t,&curLon_t, &quality); 
	
	if (valid=='A' && type[0]=='G' && type[1]=='P' && type[2]=='R' && type[3]=='M' && type[4]=='C') {
		
		
		curLatg = convertGPS(curLat_t);
		curLong = -1*convertGPS(curLon_t);

		if (!closeCoords(prevLat, prevLon, curLatg, curLong, MIN_PROG) && mode == ModeNavigator) { // if traveled far enough
			PORTD ^= 0x04;
			
		}

		if (mode == ModeExplorer) {
			//will perform actions on pushbutton input
		}

		else if (mode == ModeNavigator) {
			
			updateHeading(curLatg, curLong); //checks heading and dist to waypoint
		}
		
	}

	if (MenuState == MenuCoord) {
		char str1[6] = "Lat: \0";
		char str2[6] = "Lon: \0";

		LCDGotoXY(0,0); // display the current latitude and longitude
		LCDstring((uint8_t*)str1, 5);
		dtostrf(curLatg, 10, 6, line2); //6 decimal after
		LCDstring((uint8_t*)line2, 10);

		LCDGotoXY(0,1);
		LCDstring((uint8_t*)str2, 5);
		dtostrf(curLong, 10, 6, line2);
		LCDstring((uint8_t*)line2, 10);

	} else if (MenuState == MenuDiff) {
		if (nearWaypoint) {
			char str1[7] = "tHdg: \0";
			char str2[7] = "cHdg: \0";
		
			LCDGotoXY(0,0); // display the target heading and current heading
			LCDstring((uint8_t*)str1, 6);
			dtostrf(hdg[current_coord_index], 10, 6, line2); //6 decimal after
			LCDstring((uint8_t*)line2, 10);

			LCDGotoXY(0,1);
			LCDstring((uint8_t*)str2, 6);
			dtostrf(curHeading, 10, 6, line2);
			LCDstring((uint8_t*)line2, 10);

		} else {
			char str1[7] = "dLat: \0";
			char str2[7] = "dLon: \0";

			LCDGotoXY(0,0); // display the distance in lat and lon to the waypoint
			LCDstring((uint8_t*)str1, 6);
			dtostrf(distLat, 10, 6, line2); //6 decimal after
			LCDstring((uint8_t*)line2, 10);

			LCDGotoXY(0,1);
			LCDstring((uint8_t*)str2, 6);
			dtostrf(distLon, 10, 6, line2);
			LCDstring((uint8_t*)line2, 10);

		}
	} else if (MenuState == MenuNavig) {
		if (nearWaypoint) {
			char str1[6] = "Dir: \0";
			char str2[6] = "Deg: \0";
			char weststr[12] = "LEFT       \0";
			char eaststr[12] = "RIGHT      \0";
					
			LCDGotoXY(0,0); // display the target direction and the degree turn
			LCDstring((uint8_t*)str1, 5);
			if (directionWest) {
				LCDstring((uint8_t*)weststr, 11);
			} else {
				LCDstring((uint8_t*)eaststr, 11);
			}

			LCDGotoXY(0,1);
			LCDstring((uint8_t*)str2, 5);
			dtostrf(diffHead, 10, 6, line2);
			LCDstring((uint8_t*)line2, 10);

		} else {
			char str1[7] = "Dist: \0";
			char str2[7] = "tHdg: \0";

			LCDGotoXY(0,0); // display the distance to the waypoint and current heading
			LCDstring((uint8_t*)str1, 6);
			dtostrf(fmax(distLat,distLon), 10, 6, line2); //6 decimal after
			LCDstring((uint8_t*)line2, 10);

			LCDGotoXY(0,1);
			LCDstring((uint8_t*)str2, 6);
			dtostrf(tarHeading, 10, 6, line2);
			LCDstring((uint8_t*)line2, 10);

		}
	} else if (MenuState == MenuTransmit) {
		char str1[17] = "Push button     \0";
		char str2[17] = "to transmit: \0";

		LCDGotoXY(0,0);
		LCDstring((uint8_t*)str1, 16);
		LCDGotoXY(0,1);
		LCDstring((uint8_t*)str2, 13);
		dtostrf(explorer_coord_index,3, 0, line2);
		LCDstring((uint8_t*)line2, 3);

	} else if (MenuState == MenuMode) {
		char str1[17]    = "Mode:           \0";
		char str_nav[17] = "NAVIGATOR       \0";
		char str_exp[17] = "EXPLORER        \0";

		LCDGotoXY(0,0);
		LCDstring((uint8_t*)str1, 16);
		LCDGotoXY(0,1);
		if (mode == ModeExplorer) {
			LCDstring((uint8_t*)str_exp, 16);
		} else if (mode == ModeNavigator) {
			LCDstring((uint8_t*)str_nav, 16);
		}

	} else {
	





	//lat[current_coord_index] = 22.2222;
	//lon[current_coord_index] = 33.3333;

	LCDGotoXY(0,0);
	dtostrf(curHeading, 10, 6, line2); //6 decimal after .
//	dtostrf(debug_count, 6, 2, line2);
//	LCDstring((uint8_t*)sentence, 15);
	LCDstring((uint8_t*)line2, 10);
	//LCDstring((uint8_t*)debug_str,5);




	LCDGotoXY(0,1);
//	dtostrf(debug_count,10,6,line2);
	dtostrf(hdg[current_coord_index], 10, 6, line2);
//	LCDstring((uint8_t*)line2,10);
	LCDstring((uint8_t*)ADC_str, 10);
//	debug_count++;

	//current_coord_index += 1;
  	//if (current_coord_index == NUM_COORDS) current_coord_index = 0;

	}
  end

//
void task_menu(void) 
begin

	if (ADC_num > 230) {
		MenuState = MenuMode;
	} else if (mode == ModeNavigator) {
		if (ADC_num > 100) {
			MenuState = MenuCoord;
		} else if (ADC_num >20) {
			MenuState = MenuDiff;
		} else {
			MenuState = MenuNavig;
		}
	} else if (mode == ModeExplorer) {
		if (ADC_num >20) {
			MenuState = MenuCoord;
		} else {
			MenuState = MenuTransmit;
		}
	}

end

void task_adc(void) 
begin  
	ADC_num = ADCH; //get the sample of ADC
 	ADCSRA |= (1<<ADSC) ; //start another conversion
 	dtostrf(ADC_num, 4, 0, ADC_str);
 	
end

int compare(char a[], char b[])
{
   int c = 0;
 
   while( a[c] == b[c] )
   {
      if( a[c] == '\0' || b[c] == '\0' )
         break;
      c++;
   }
   if( a[c] == '\0' && b[c] == '\0' )
      return 0;
   else
      return -1;
}

float convertGPS(float num_gps)
{
   int degrees = (int)num_gps / 100;
   float minutes = (num_gps - 100*degrees) / (float)60;
   return degrees+minutes;
}

char closeCoords(float target_lat, float target_lon,  float cur_lat, float cur_lon, float dist) {
	float distance_lat = fabsf(target_lat - cur_lat);
	float distance_lon = fabsf(target_lon - cur_lon);

	if (distance_lat < dist) {
		if (distance_lon < dist) {
			return 1;
		}
	}
	return 0;

}

void stringcopy(char *dest, char *src, int beg_index, int end_index) {
	int i;
	int j=0;
	for (i=beg_index; i<end_index; i++) {
		dest[j] = src[i];
		j++;
	}
	dest[j] = '\0';

}

void sendToComputer(void) {
	int i;
	
	for (i=0; i<explorer_coord_index; i++) {
		char latstr[11];
		char lonstr[11];

		dtostrf(lat_exp[i], 10, 6, latstr);
		dtostrf(lon_exp[i], 10, 6, lonstr);

		

		fprintf(&uart_str,"lat[%d] = %s; ", i, latstr);	
		fprintf(&uart_str,"lon[%d] = %s;\n", i, lonstr);	
	
	}
}


void receive_gps_uart(char c){



		//UDR0 = c;	// Print raw characters to PC (used for testing)
		char rx_c_bit = c;
		//char rx_c_bit = uart_getchar(stdin);
		rx_c[rx_size] = rx_c_bit;
		if (rx_size <1024) {rx_size++;}
		if (rx_c_bit == '$') {
			//fprintf(stdout,rx_c, rx_size);
		
		
			stringcopy(sentence,rx_c,0,rx_size);
		
			sentence_len = rx_size;
			rx_size = 0;

		}


	

}

//*******************************   
//Task 3  
void button(void) 
begin

  switch (PushState)
  begin
     case NoPush: 
        if (~PINB & 0x04) PushState=MaybePush;
        else PushState=NoPush;
        break;
     case MaybePush:
        if (~PINB & 0x04) 
        begin
           PushState=Pushed;  
		   PORTD ^= 0x04;
		   // button was pushed
			if (MenuState == MenuMode) {
				if (mode == ModeExplorer) { mode = ModeNavigator; }
				else { mode = ModeExplorer; nearWaypoint=0;}
			} else if (MenuState == MenuTransmit) {
				sendToComputer();
				explorer_coord_index = 0;
			} else if ((MenuState == MenuCoord) && (mode == ModeExplorer)) {
				if (explorer_coord_index < NUM_EXPLORER_WAYPTS) {
				

					lat_exp[explorer_coord_index] = curLatg;
					lon_exp[explorer_coord_index] = curLong;
					explorer_coord_index++;
				}
			}

		   
        end
        else PushState=NoPush;
        break;
     case Pushed:  
        if (~PINB & 0x04) {PushState=Pushed;
		} 
        else PushState=MaybeNoPush;    
        break;
     case MaybeNoPush:
        if (~PINB & 0x04) PushState=Pushed; 
        else 
        begin
           PushState=NoPush;
        end    
        break;
  end
end


void updateHeading(float curLat, float curLon) {



	distLat = fabsf(lat[current_coord_index] - curLat);
	distLon = fabsf(lon[current_coord_index] - curLon);
	

	debug_count = fmax(distLat,distLon);

	// check distance from previous coord
	if (!closeCoords(prevLat, prevLon, curLat, curLon, MIN_PROG)) { // if traveled far enough
		
		float diffLat_t = lat[current_coord_index] - curLat;
		float diffLon_t = lon[current_coord_index] - curLon;
		tarHeading = 90 - atan2(diffLat_t,diffLon_t)*180/PI;
		if (tarHeading < 0) {tarHeading=tarHeading+360;}

		float diffLat = curLat - prevLat;
		float diffLon = curLon - prevLon;
		newHeading = 90 - atan2(diffLat,diffLon)*180/PI;
		if (newHeading < 0) {newHeading=newHeading+360;}

		// update heading: forgetful averaging
		if ( fabsf(newHeading - curHeading) > 180 ) { // want to avg 1deg and 359deg to 0deg not 180deg
			
			if (newHeading < curHeading) {newHeading = newHeading+360; }
			else {curHeading = curHeading + 360;}
			curHeading = forget*curHeading + (1-forget)*newHeading;
		} else {
			curHeading = forget*curHeading + (1-forget)*newHeading;
		}
		if (curHeading >= 360) {curHeading = curHeading - 360;}

		// save prev coord
		prevLat = curLat;
		prevLon = curLon;
	}

	
	if (closeCoords(lat[current_coord_index], lon[current_coord_index], curLat, curLon, MIN_DIST_WAY)) { // if close enough to next waypoint
		nearWaypoint = 1;
	}

	if (nearWaypoint) {
		diffHead = fabsf(hdg[current_coord_index] - curHeading);

		if ( diffHead > 180 ) {	
			diffHead = 360 - diffHead;
			directionWest = hdg[current_coord_index] > curHeading ? 1 : 0;
		} else {
			directionWest = hdg[current_coord_index] > curHeading ? 0 : 1;
		}

		// led blink (length, chirp #)
		float dist = fmax(distLat,distLon);
		

		syll_length= 100 + 150 * (dist/MIN_DIST_WAY);
		num_syll = (int)((diffHead+22.5)/45) + 1;

		
		// if heading matches destination
		

		if (diffHead < MIN_HEAD_DIFF) {
			current_coord_index++;	// proceed to next waypoint
			current_coord_index %= NUM_COORDS;
			nearWaypoint = 0;
			chirp_rpt_count = 0;
			num_syll_count = 0;
			syll_length_count = 0;
			OCR1A = -1; //turn off output
			OCR1B = -1;
		}
	}
}

void ledblink(void) {
	chirp_rpt_count +=1;

	// start a new chirp
	if (chirp_rpt_count >= CHIRP_RPT) {
		chirp_rpt_count = 0;
		num_syll_count = 0;
		syll_length_count = 0;
	}



	if (syll_length_count > syll_length*2) {
		
		if (num_syll_count+1 < num_syll) {
			num_syll_count +=1;
			syll_length_count = 0;
		}

	} else if (syll_length_count > syll_length) {
		syll_length_count +=1;
		OCR1A = -1; //turn off output
		OCR1B = -1;
	} else { // syll_length_count <= syll_length
		syll_length_count +=1;
		if (directionWest > 0) {
			OCR1A = 0; //turn on output
		} else {
			OCR1B = 0;
		}
	}
}

























  
//**********************************************************
// LCD setup
void init_lcd(void) 
begin
	char line1[11] = "GPS Coord:\0";

	LCDinit();	//initialize the display
	LCDcursorOFF();
	LCDclr();	//clear the display
	
	LCDGotoXY(0,0);
	LCDstring((uint8_t*)line1, 10);

	

end



//**********************************************************       
//Entry point and task scheduler loop
int main(void)
begin  
  initialize();
  init_coords();
  
  // main task scheduler loop 
  while(1)
  begin
  
  	// check if tasks are ready    
    if (count_lcd==0){count_lcd=time_lcd;	LCD_print();}
	if (count_adc==0){count_adc=time_adc;	task_adc();}
    if (count_led==0){count_led=time_led;	task_menu();}
	if (count_but==0){count_but=time_but;   button();}
	

  end
end  
  

//Set it all up
void initialize(void)
begin

  DDRD = 0x00;	// PORT D is an input 
  DDRB = 0x00;  // for button
          
  // led blink outputs
  DDRD |= 0x30;
  TCCR1B = 1;  
  TCCR1A = (1<<COM1A0) | (1<<COM1A1) | (1<<COM1B0) | (1<<COM1B1) | (1<<WGM10) | (1<<WGM11) ; 
  OCR1A = -1; //turn off output
  OCR1B = -1;

  //set up timer 0 for 1 mSec timebase 
  TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR 
  OCR0A = 249;  		//set the compare re to 250 time ticks
  TCCR0B= 3;    //0b00000011;  //set prescalar to divide by 64	
  TCCR0A= (1<<WGM01); // turn on clear-on-match
  
  // setup LED
  DDRD |= 0x04;
  PORTD ^= 0x04;

  // setup LCD
  init_lcd();

  //init UART
  uart1_init(receive_gps_uart);
  stdout = stdin = stderr; //= &uart_str;
  fprintf(stdout,"$PMTK220,1000*1F");
  fprintf(stdout,"$PGCMD,33,1*6C");
  fprintf(stdout,"$PMTK314,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*28");
  

  //init the A to D converter 
  ADMUX = (1<<REFS0) | (1<<ADLAR) ;    
  //enable ADC and set prescaler to 1/128*16MHz=125,000
  //and clear interupt enable and start a conversion
  ADCSRA = ((1<<ADEN) | (1<<ADSC)) + 7 ; 
  
  //init the task timers
  count_lcd=time_lcd;
  count_led=time_led;
  count_adc=time_adc;
  count_but=time_but;
      
  //crank up the ISRs
  sei();

end 



















